import numpy as np
from scipy.special import wofz
import matplotlib.pyplot as plt
import astropy.units as u
import astropy.constants as c

def voigt(x, dx, sigma, gamma, amplitude):
    """Generate a Voigt profile at positions x, given the parameters sigma and gamma.

    Args: 
        x (arr): given to generate the Voigt profile
        dx (float): location of the center of the profile
        sigma(float): parameter characterizing the Voigt profile
        gamma(float): parameter characterizing the Voigt profile
        amplitude(float): amplitude of the Voigt profile
    Returns:
        v (arr): y values of the Voigt profile
        
    """
    z = ((x-dx) + 1j*gamma) / (sigma * np.sqrt(2))
    v = amplitude*np.real(wofz(z)) / (sigma * np.sqrt(2*np.pi))
    return v

def generate_voigt_profile(x,dx, sigma, gamma, amplitude, noise_std=0):
    """Generate a Voigt profile with given parameters at positions x, including Gaussian noise.
    
    Args: 
        x (arr): given to generate the Voigt profile
        dx (float): location of the center of the profile
        sigma(float): parameter characterizing the Voigt profile
        gamma(float): parameter characterizing the Voigt profile
        amplitude(float): amplitude of the Voigt profile
        noise_std(float): standard deviation of noise to be applied to Voigt profile
                        default is 0
    
    Returns:
        y_with_noise (arr): Voigt profile with Gaussian noise

    """
    y = voigt(x, dx, sigma, gamma, amplitude)
    noise = np.random.normal(0, noise_std, len(x))
    y_with_noise = y + noise
    return y_with_noise

#Example
#------------------------------
# Generate x values
x = np.linspace(-10, 10, 1000)
dx = 0
# Set initial parameter values
sigma = 1.0
gamma = 1.0
amplitude = 100.0
noise_std = 0.1  # Standard deviation of the Gaussian noise

# Generate Voigt profile with noise
y = generate_voigt_profile(x,dx, sigma, gamma, amplitude, noise_std)

# Plot the Voigt profile with noise
plt.plot(x, y)
plt.xlabel('x')
plt.ylabel('Intensity')
plt.title('Voigt Profile with Noise')
plt.show()

#Saving to a .csv file
# data = np.column_stack((x, y))
# np.savetxt('voigt_data.csv', data, delimiter=',', header='x,y', fmt='%.6f', comments='')



### from https://academic.oup.com/mnras/article/458/2/1427/2589008
def gamma_g(temp, species, v0):
    """
    Return the width of the Gaussian component of the Voigt profile.

    Parameters:
        temp (float): Temperature of planetary atmosphere in K.
        species (str): 'H2O' or 'CO2' -- species from which to calculate the pressure.
        v0 (float): Central wavenumber of line -- in 1/cm.

    Returns:
        gamma_g (float): Width of Gaussian component of Voigt profile in 1/cm.
    """
    if species.casefold() == 'h2o':
        mass = (2*c.m_p +  8*c.m_p).to(u.g)
    if species.casefold() == 'co2':
        mass = (6*c.m_p + 16*c.m_p).to(u.g)

    if species.caseefold() not in ['h2o', 'co2']:
        raise Exception('Specify chemical species -- "H2O" or "CO2"')
    
    gamma_g = np.sqrt(2*c.k_B*temp*u.K/mass) * (v0/u.cm)/c.c.to(u.cm/u.s)).to(1/u.cm)

    return gamma_g.value

def gamma_l(temp, press, species):
    """
    Return the width of the Lorentzian component of the Voigt profile.

    Parameters:
        temp (float): Temperature of planetary atmosphere in K.
        press (float): Pressure in kPa.
        species (str): 'H2O' or 'CO2' -- species from which to calculate the pressure.

    Returns:
        gamma_l (float): Width of Lorentzian component of Voigt profile in 1/cm.
    """
    if species.casefold() == 'h2o':
        n = 0.58 ## mean for broadening due to air across H_2^16O transitions from HITRAN
        gl_mol = 0.068 ## mean for broadening due to air across H_2^16O transitions from HITRAN
    if species.casefold() == 'co2':
        n = 0.71 ## mean for broadening due to air across ^12C^16O_2 transitions from HITRAN
        gl_mol = 0.070 ## mean for broadening due to air across ^12C^16O_2 transitions from HITRAN
    if species.caseefold() not in ['h2o', 'co2']:
        raise Exception('Specify chemical species -- "H2O" or "CO2"')

    partial_press = 1 #partial pressure, assuming only one species and isotopologue, so = 1

    gamma_l = ((temp*u.K)/(296*u.K))**n * press*u.kPa.to(u.atm) * (gl_mol/(u.cm*u.atm) * partial_press) #would sum, but only one species

    return gamma_l.to(1/u.cm).value

def gamma_voigt(temp, press, species, v0):
    """
    Return the Voigt profile width based on Gaussian and Lorentzian components.

    Parameters:
        temp (float): Temperature of planetary atmosphere in K.
        press (float): Pressure in kPa.
        species (str): 'H2O' or 'CO2' -- species from which to calculate the pressure.
        v0 (float): Central wavenumber of line -- in 1/cm.
    
    Returns:
        gamma_voigt (float): Width of Voigt profile in 1/cm.
    """
    gg = gamma_g()
    gl = gamma_l()

    return 0.5346*gl + np.sqrt(0.2166*gl**2 + gg**2)


from scipy.optimize import curve_fit

## with some def voigt based on what you've already written

# def multi_voigt(x, sigma, gamma, amplitude, noise_std, nfit):
def multi_voigt(x, sigma, gamma, amplitude, noise_std, nfit):
    ## should work for nfit
    ## sigma, gamma, amplitude, noise will have to be lists


    for i in range(nfit):


def wrapper_fit_func(x, N, *args):
    a, b, c = list(args[0][:N]), list(args[0][N:2*N]), list(args[0][2*N:3*N])
    return fit_func(x, a, b, c, N)

def multi_voigt_wrapper()


def voigt_fit(voigt, wave, spec, nfit, **kwargs):

    popt, pcov = curve_fit(lambda x, sigma, gamma, amplitude, noise_std: multi_voigt(x, sigma, gamma, amplitude, noise_std, nfit), 
                           wave, spec, **kwargs)
    
    ## reorder popt to be sigma, gamma, amplitude, noise_std
    len_ = len(popt)/4
    sigma, gamma, amplitude, noise_std = popt[:len_], popt[len_:2*len_], popt[2*len_:3*len_], popt[3*len_:]
    
    return sigma, gamma, amplitude, noise_std


from teltrace.load import file_read
import matplotlib.pyplot as plt

filepath = "/Users/bensappey/Downloads/Source_file_METEC_04NOV_2021_methane_background.csv"

spectrum = file_read(filepath=filepath)
print(spectrum)

# plt.plot()
# plt.show()


import numpy as np
import pandas as pd
import astropy.io.fits as fits


def file_read(filepath):
    """Read a .csv file into a dataframe to be manipulated

    Args:
        filepath: (str, path object, or file-like object) .csv file to be read in

    Returns:
        spectrum (pandas.df) Dataframe containing wavenumber information
        in the first column and intensity of line information in the second column
    """
    #load a .csv file and output a dataframe with wavelength and intensity information
    spectrum  = pd.read_csv(filepath, header=[0])
    #rename columns
    spectrum.columns.values[0] = 'Wavenumber'
    spectrum.columns.values[1] = 'Intensity'
    return spectrum

def fits_read(filepath, wavekey = 'wave', speckey = 'spec'):
    """
    Read in spectrum from a FITS file containing a table.

    Parameters:
        filepath (str): Path to FITS file to be read in.
        wavekey (str): Key corresponding to wavelength information in the FITS file.
        speckey (str): Key corresponding to intensity information in the FITS file.

    Returns:
        wave (arr): Wavelengths axis from FITS file.
        spec (arr): Intensity axis from FITS file.
    """

    dat = fits.open(filepath)[0].data

    wave, spec = dat[wavekey], dat[speckey]

    return wave, spec



## still haven't found a database, so placeholder code in the meantime

import pandas as pd
import matplotlib.pyplot as plt
import numpy as np

def get_pressure(temp, species = 'h2o'):
    """
    Return pressure based on pressure-temperature curve for H2O or CO2.

    Parameters:
        temp (float): Temperature of planetary atmosphere in K.
        species (str): 'H2O' or 'CO2' -- species from which to calculate the pressure.
    
    Returns:
        pressure (float): Pressure in kPa.
    """
    if species.casefold() == 'h2o':

    if species.casefold() == 'co2':

    if species.caseefold() not in ['h2o', 'co2']:
        raise Exception('Specify chemical species -- "H2O" or "CO2"')

def object_profile(name, species = 'h2o',
                   figsize = (5, 5), color = 'gray', ls = '-', 
                   xlabel = 'x', ylabel = 'intensity', 
                   xscale = 'linear', yscale = 'linear'):
    """
    Return Voigt profile of line based on conditions of specific planet.

    Parameters:
        name (str): Name of planet (without spaces).
        species (str): 'H2O' or 'CO2' -- species from which to calculate the pressure.
            Passed to get_pressure().

        See matplotlib documentation for following --
            figsize (tuple): Size of plotted figure.
            color (str): Color of plotted profile.
            ls (str): Line style of plotted profile.
            xlabel (str): Label on x-axis of plotted figure.
            ylabel (str): Label on y-axis of plotted figure.
            xscale (str): Scaling of x-axis of plotted figure.
            yscale (str): Scaling of y-axis of plotted figure.
    
    Returns:
        Plot of line profile. 
    """

    props = pd.read_csv('PLACEHOLDER')

    if name.casefold() not in props.names.str.casefold():
        raise Exception('Planet conditions not found. List available planets with list_object.')
    
    if name.casefold() in props.names.str.casefold():
        planet_props = props.loc[props.names.str.casefold() == name.casefold()]

        press = get_pressure(planet_props.temp, species)

        x, spec = PLACEHOLDER_VOIGT_FUNCTION(planet_props.temp, press)

        fig = plt.figure(figsize = figsize)
        plt.plot(x, spec, color = color, ls = ls)
        plt.xlabel(xlabel)
        plt.ylabel(ylabel)
        plt.xscale(xscale)
        plt.yscale(yscale)

    

def list_object():
    props = pd.read_csv('PLACEHOLDER')
    for i in props.names.str.casefold():
        print(i)


import matplotlib.pyplot as plt
import numpy as np
from scipy.optimize import curve_fit
from teltrace.gen_voigt import voigt, generate_voigt_profile

def fit_voigt(x, y, bounds=None):

    """Function to fit a Voigt profile to a line

    Args:
        x: (array) wavenumber data
        y: (array) intensity data
        bounds: 

    Returns:
        Best fit parameters  (array), containing the 
        best fit parameters of the Voigt profile
        (dx, sigma, gamma, amplitude).
        pcov (2D-array), estimated approximate covariance of popt 
    """

    if bounds is None:
        popt, pcov = curve_fit(xdata=x, ydata=y,
                               f=voigt,
                               p0=[x[np.argmax(y)], None, None, 2*np.max(y)])
        return popt,pcov
    else:
        popt,pcov =  curve_fit(xdata=x,ydata=y,f = voigt,
                               p0=[x[np.argmax(y)], None, None, 2*np.max(y)],
                               bounds=bounds)
        return popt,pcov

def plot_fit_voigt(x,y,bounds=None):
    """Function that plots the best-fit Voigt Profile against data

    Args:
        x: (array) wavenumber data
        y: (array) intensity data
        bounds:
    
    Returns:
        Best fit parameters  (array), containing the 
        best fit parameters of the Voigt profile
        (dx, sigma, gamma, amplitude).
        plt:(matplotlib object) plot the data and best fit curve together
    """
    popt,pcov = fit_voigt(x,y,bounds=bounds)

    y = generate_voigt_profile(x, dx=popt[0], sigma=popt[1], gamma=popt[2], amplitude=popt[3], noise_std=0)

    # Plot the Voigt profile from parameters
    plt.plot(x, y)
    plt.xlabel('x')
    plt.ylabel('Intensity')
    plt.title('Voigt Profile from Parameters')
    plt.show()
    return popt, plt


import os

__version__ = "0.1"

# set Python env variable to keep track of example data dir
teltrace_dir = os.path.dirname(__file__)
DATADIR = os.path.join(teltrace_dir, "data/")


try:
    from . import load

    cext = True
except ImportError:
    cext = False
